
//*******************************************************************
//
//  UNIT:     G r 3 2 _ G o u r a u d       05.11.2003,  27.06.2005
//
//  Extended graphic unit for graphics32 library
//  Additional drawing routines with gouraud shading
//
//  Copyright (c) 2005 by Gerd Platl
//
//  --- functions and procedures ---
//
//  SetMasterAlpha      set master opacity
//  GouraudHLine        draw a horizontal gouraud line
//  GouraudTriangle     draw a filled gouraud triangle
//  GouraudRectangle    draw a filled gouraud rectangle
//  DrawTriangle        draw triangle lines
//  DrawRGBColors       draw RGB colors in rectangle
//
//*******************************************************************
unit Gr32_Gouraud;

{$I GR32.INC}

{$ifopt D+}
  {$define xtest1   main function test}
  {$define xtest2   sub function test}
{$endif}

INTERFACE

uses  Windows, Gr32, Gr32_Blend;

type  p2i = record                     // 2d integer point
        x,y: longint;   end;

type  tTriangle = record               // triangle type
        a,b,c: p2i;   end;             // 3 vertices

type  tColor3v = record                // triangle vertex colors
        a,b,c: tColor32;   end;        // 3 colors

procedure SetMasterAlpha (alpha: byte);

procedure GouraudHLine (x1,x2,y: integer; c1,c2: tColor32);
procedure GouraudTriangle (tp: tTriangle; tc: tColor3v);
procedure GouraudRectangle (x1,y1,x2,y2: integer;
                            c1,c2,c3,c4: tColor32);

procedure DrawTriangle (tp: tTriangle; color: tColor32);
procedure DrawRGBColors (left, top, xdim,ydim: integer;
                         alpha: byte = 255);

//*********************************************************
IMPLEMENTATION

uses  SysUtil2, Gr32_Ext;

var   hLine32Bm: tBitmap32 = nil;      // 32 bit work line
      _masterAlpha: byte = 255;        // master opacity

//---------------------------------------------------------
procedure IncColor  (var c1,c2: tLongColor);  // c1 = c1 + c2;
begin
  inc (c1.a, c2.a);
  inc (c1.r, c2.r);
  inc (c1.g, c2.g);
  inc (c1.b, c2.b);
end;
{*********************************************************}
{ set master opacity
{ input:     alpha: byte;  color opacity (0..255)
{---------------------------------------------------------}
procedure SetMasterAlpha (alpha: byte);
begin
  _masterAlpha := alpha;
//  activeBitmap.MasterAlpha := alpha;   //???
end;

//*********************************************************
// convert color32 values to 24.8 fix point values
//---------------------------------------------------------
function LongColor (color: tColor32): tLongColor;
begin
  with tColorRGBA(color) do begin
    result.a := a shl 8;
    result.r := r shl 8;
    result.g := g shl 8;
    result.b := b shl 8;
  end;
end;
//*********************************************************
// convert 24.8 fixpoint color to tColor32
//---------------------------------------------------------
function Color32 (color: tLongColor): tColor32;
begin
  result := longword(color.a) shl 16 and $FF000000
           +longword(color.b) shl 8  and $00FF0000
           +longword(color.g)        and $0000FF00
           +longword(color.r) shr 8  and $000000FF;
end;

//*********************************************************
// draw triangle lines using given color
//---------------------------------------------------------
procedure DrawTriangle (tp: tTriangle; color: tColor32);
begin
  with tp.a do activeBitmap.LineTS (x,y, tp.b.x, tp.b.y, color);
  with tp.b do activeBitmap.LineTS (x,y, tp.c.x, tp.c.y, color);
  with tp.c do activeBitmap.LineTS (x,y, tp.a.x, tp.a.y, color);
end;

//*********************************************************
// Calculate color start value c1 and color step value c2.
// If v1 < 0 then c1 will be corrected and set v1 = 0.
// input:     v1,v2  start and end value of line
//            c1,c2  start and end value of color
// output:    v1     corrected start value
//            c1,c2  start and step value of color
// return:   boolean =false nothing to calculate
//---------------------------------------------------------
function ColorStartAndStep (var v1,v2: integer;
                            var c1,c2: tLongColor): boolean;
var vd: integer;
begin
  result := false;
  vd := v2 - v1;
  c2.r := c2.r - c1.r;                 // c2 = c2 - c1
  c2.g := c2.g - c1.g;
  c2.b := c2.b - c1.b;
  c2.a := c2.a - c1.a;
  if vd <= 0 then exit;                // if start <= end value ===>
  if v2 <  0 then exit;                // ===>
  if v1 < 0
  then begin                           // clip left side
    c1.r := c1.r - c2.r * v1 div vd;   // calculate color start values
    c1.g := c1.g - c2.g * v1 div vd;
    c1.b := c1.b - c2.b * v1 div vd;
    c1.a := c1.a - c2.a * v1 div vd;
    v1 := 0;
  end;
  c2.r := c2.r div vd;                 // calculate color step values
  c2.g := c2.g div vd;
  c2.b := c2.b div vd;
  c2.a := c2.a div vd;
  result := true;
end;

//*********************************************************
// draw horizontal line from point x1/y to x2/y,
// where color values are interpolated linear from c1 to c2
// note:   x1 must be less or equal x2.
//---------------------------------------------------------
procedure GouraudHLineLong (x1,x2,y: integer; c1,c2: tLongColor);

var   xi, maxX,maxY: longint;
      p1: PColor32;           // pixel color pointer
begin
  maxX := activeBitmap.Width - 1;
  maxY := activeBitmap.Height - 1;

{$ifdef test2}
  writeln (tst, 'GouraudHLine: ', x1:4,x2:4,y:4, ' a=',_masterAlpha:4
              , HexString(Color32(c1)):10, HexString(Color32(c2)):10);
//  writeln (tst, '   maxX=', maxX:5, '  maxY=',maxY);
  flush (tst);
{$endif}

  if (y < 0) or (y > maxY)               then exit;  // ===>
  if x1 > x2                             then exit;  // ===>

  if _masterAlpha < 255 then begin
    c1.a := integer(c1.a) * _masterAlpha div 255;
    c2.a := integer(c2.a) * _masterAlpha div 255;
  end;

  if not ColorStartAndStep (x1,x2,c1,c2) then exit;  // ===>
  if x2 > maxX then x2 := maxX;

  try
    p1 := activeBitmap.PixelPtr[x1,y];
    for xi := x1 to x2
    do begin
      BlendMem (Color32(c1), p1^);     // put transparent pixel
      inc (p1);                        // inc destination pointer
      inc (c1.r, c2.r);                // IncColor (c1,c2);
      inc (c1.g, c2.g);
      inc (c1.b, c2.b);
      inc (c1.a, c2.a);
    end;
  finally
    EMMS;
  end;
end;
//*********************************************************
// draw horizontal gradient line from point x1/y to x2/y,
// where color values are interpolated linear from c1 to c2
// note:   x2 must be greater then x1.
// input:  x1,y         line starting point
//         x2,y         line end point
//         c1,c2        starting and ending color
//---------------------------------------------------------
procedure GouraudHLine2 (x1,x2,y: integer; c1,c2: tColor32);
begin
  GouraudHLineLong (x1,x2,y, LongColor(c1), LongColor(c2));
end;
//*********************************************************
// draw horizontal gradient line from point x1/y to x2/y,
// where color values are interpolated linear from c1 to c2
// note:   x2 must be greater then x1.
// adapted from example GradLines_Ex.exe
//---------------------------------------------------------
procedure GouraudHLine (x1,x2,y: integer; c1,c2: tColor32);
begin
  if x1 > x2 then Exit;
  if x1 = x2 then activeBitmap.SetPixelT (x1,y, c1)
  else begin
    activeBitmap.SetStipple([c1, c2]);
    activeBitmap.StippleStep := 1 / (x2-x1);
    activeBitmap.StippleCounter := 0;
    activeBitmap.HorzLineTSP(x1, y, x2);
  end;
end;

//*********************************************************
// draw a filled gouraud rectangle (x1,y1 .. x2,y2)
// with interpolated RGB-Values (c1-c2 .. c3-c4):
//    c1+========+c2
//      |        |
//   cy1+..+cx...+cy2
//      |        |
//    c3+========+c4
//---------------------------------------------------------
procedure GouraudRectangle (x1,y1,x2,y2: integer;
                            c1,c2,c3,c4: tColor32);
var  yi: longint;
     cl1,cl2,cl3,cl4: tLongColor;
begin

{$ifdef test1}
  writeln (tst, 'GouraudRectangle: ', x1:4,y1:4, x2:4,y2:4
              , HexString(c1):10, HexString(c2):10
              , HexString(c3):10, HexString(c4):10);
{$endif}

  yi := y1;                    // save y1 for negative values
  cl1 := LongColor (c1);
  cl2 := LongColor (c2);
  cl3 := LongColor (c3);
  cl4 := LongColor (c4);
  if not ColorStartAndStep (y1,y2,cl1,cl3) then exit;  // ===>
  if not ColorStartAndStep (yi,y2,cl2,cl4) then exit;  // ===>
  for yi := y1 to y2
  do begin
    GouraudHLineLong (x1,x2, yi, cl1,cl2);
    IncColor (cl1, cl3);
    IncColor (cl2, cl4);
  end;
end;
//---------------------------------------------------------
//  draw gouraud rectangle with calling GouraudHLine2
//---------------------------------------------------------
procedure GouraudRectangle2 (x1,y1,x2,y2: integer;
                            c1,c2,c3,c4: tColor32);
var  yi: longint;
     cl1,cl2,cl3,cl4: tLongColor;
begin

{$ifdef test1}
  writeln (tst, 'GouraudRectangle: ', x1:4,y1:4, x2:4,y2:4
              , HexString(c1):10, HexString(c2):10
              , HexString(c3):10, HexString(c4):10);
{$endif}

  yi := y1;                    // save y1 for negative values
  cl1 := LongColor (c1);
  cl2 := LongColor (c2);
  cl3 := LongColor (c3);
  cl4 := LongColor (c4);
  if not ColorStartAndStep (y1,y2,cl1,cl3) then exit;  // ===>
  if not ColorStartAndStep (yi,y2,cl2,cl4) then exit;  // ===>
  for yi := y1 to y2
  do begin
    GouraudHLine2 (x1,x2, yi, Color32(cl1), Color32(cl2));
    IncColor (cl1, cl3);
    IncColor (cl2, cl4);
  end;
end;

//*********************************************************
// draw color table at position (left,top)       07.06.2005
// as a rectangle (xdim*ydim) with rgb values.
// transpoarent value alpha can be set optionell.
//
// left/top
//   +================== B l a c k ==================+
//   |       |       |       |       |       |       |
//   |       |       |       |       |       |       |
// Magenta--RED----Yellow--GREEN---Cyan----BLUE--Magenta
//   |       |       |       |       |       |       |
//   |       |       |       |       |       |       |
//   +================== W h i t e ==================+
//---------------------------------------------------------
procedure DrawRGBColors (left, top, xdim,ydim: integer;
                         alpha: byte = 255);

type  tColors = array[0..8] of tColor32;
const baseColors: tColors
      = (clBlack32, clFuchsia32, clRed32, clYellow32, clLime32
        ,clAqua32, clBlue32, clFuchsia32, clWhite32);

var   x1,x2, y1,y2, ri: word;    c1,c2: tColor32;   colors: tColors;
begin
  y1 := top+ydim div 2;
  y2 := top+ydim-1;
  for ri := 0 to 8
  do colors[ri] := baseColors[ri] and $00FFFFFF + alpha shl 24;
  for ri := 0 to 5
  do begin
    c1 := colors[ri+1];
    c2 := colors[ri+2];
    x1 := left + (ri*xdim) div 6;
    x2 := left + ((ri+1)*xdim) div 6 - 1;
    // black .. rgb
    GouraudRectangle (x1, top, x2, y1-1, colors[0], colors[0], c1,c2);
    // rgb .. white
    GouraudRectangle (x1, y1, x2, y2, c1,c2, colors[8], colors[8]);
  end;
end;

//*********************************************************
//
//        G O U R A U D    -    T R I A N G L E
//
//---------------------------------------------------------
// cases after sorting triangle points by y value
//         1. a-----b   2. a         3.  a      4.  a
//             \   /      / \           / \        / \
//              \ /      b.  \         /  .b      /   \
//               c          . \       / .        b-----c
//                            .c     c.
//---------------------------------------------------------

//*********************************************************
// draw Gouraud shaded triangle
// input:     tp:   vertices
//            tc:   colors of vertices
//---------------------------------------------------------
procedure GouraudTriangle (tp: tTriangle; tc: tColor3v);

  //*******************************************************
  // sort triangle vertices and colors by y value
  // and if y values are equal by x value
  //-------------------------------------------------------
  procedure SortTP (var tp: tTriangle; var tc: tColor3v);

    //-----------------------------------------------------
    // swap vertices and colors
    //-----------------------------------------------------
    procedure SwapVC (var p1,p2: p2i; var c1,c2: tColor32);
    (*
    var   ph: p2i;   ch: tColor32;
    begin
      ph := p1;  ch := c1;
      p1 := p2;  c1 := c2;
      p2 := ph;  c2 := ch;
    *)
    begin                    // fast solution
      SwapMem (p1,p2, sizeof(p2i));
      SwapMem (c1,c2, sizeof(tColor32));
    end;

  begin            //--- SortTP ---
    with tp
    do begin
      // sort point by y values
      if a.y > b.y then SwapVC (a,b, tc.a, tc.b);
      if b.y > c.y then SwapVC (b,c, tc.b, tc.c);
      if a.y > b.y then SwapVC (a,b, tc.a, tc.b);

      // if y values are equal, sort by x values
      if a.y = b.y
      then if a.x > b.x    // case 1 ?
        then SwapVC (a,b, tc.a, tc.b);

      if b.y = c.y
      then if b.x > c.x    // case 4 ?
        then SwapVC (b,c, tc.b, tc.c);
    end;
  end;
  //*******************************************************
  // calculate step value using 24.8 fixpoint notation
  // Versorgung:  a1,a2: longint   start and end value
  //              d: longint       number of steps
  //-------------------------------------------------------
  function Delta (a1,a2, d: longint): longint;
  begin
    if d = 0
    then result := 0
    else result := (a2-a1) * 256 div d;
  end;
  //*******************************************************
  // calculate start and step values as 24.8 fixpoint values
  // input:    p1,p2    line point start & end value
  //           c1,c2    color start & end value
  // output:   start values   sx, sc
  //           step values    dx, dc
  //-------------------------------------------------------
  procedure StartAndStep (const p1, p2: p2i;
                          const c1, c2: tColor32;
                          var sx, dx: longint;
                          var sc, dc: tLongColor);
  var   yd: longint;
  begin
{$ifdef test2}
  writeln (tst, '  StartAndStep: ', p1.x:4,p1.y:4, p2.x:6,p2.y:4);
{$endif}

    sx := longint(p1.x) * 256;           // start value X
    yd := p2.y - p1.y;                   // y steps
    dx := Delta (p1.x, p2.x, yd);        // step value X
    sc := LongColor (c1);                // start value color
                                         // step values color
    dc.r := Delta (tColorRGBA(c1).r, tColorRGBA(c2).r, yd);
    dc.g := Delta (tColorRGBA(c1).g, tColorRGBA(c2).g, yd);
    dc.b := Delta (tColorRGBA(c1).b, tColorRGBA(c2).b, yd);
    dc.a := Delta (tColorRGBA(c1).a, tColorRGBA(c2).a, yd);
  end;

  //*******************************************************
  var   x1,x2, d1,d2,d3: longint;
        c1,c2, f1,f2,f3: tLongColor;
  //-------------------------------------------------------
  // draw triangle part using fixpoint
  // input:   ys,ye:    y start & end value
  //-------------------------------------------------------
  procedure FillPartFix (ys, ye: integer);

  var   yi: integer;
  begin
    if ye > activeBitmap.width then ye := activeBitmap.width;
    for yi := ys to ye
    do begin
      if yi >= 0
      then GouraudHLineLong (x1 div 256, x2 div 256, yi, c1,c2);
      inc (x1, d1);                              // next start position
      inc (x2, d2);                              // next end position
      inc (c1.r, f1.r);   inc (c2.r, f2.r);      // next color values
      inc (c1.g, f1.g);   inc (c2.g, f2.g);
      inc (c1.b, f1.b);   inc (c2.b, f2.b);
      inc (c1.a, f1.a);   inc (c2.a, f2.a);
    end;
  end;

begin              //--- GouraudTriangle ---
{$ifdef test1}
  writeln (tst,'Gr32_Gouraud: GouraudTriangle: ', tp.a.x:5, tp.a.y:5);
{$endif}

  // DrawTriangle (tp, tc.a);  {for testing only}

  SortTP (tp, tc);

  with tp do begin
    if a.y = b.y
    then begin               // case 1
      StartAndStep (a,c, tc.a, tc.c, x1,d1, c1,f1);
      StartAndStep (b,c, tc.b, tc.c, x2,d2, c2,f2);
      FillPartFix (a.y, c.y);
      exit;
    end
    else if b.y = c.y
    then begin               // case 4
      StartAndStep (a,b, tc.a, tc.b, x1,d1, c1,f1);
      StartAndStep (a,c, tc.a, tc.c, x2,d2, c2,f2);
      FillPartFix (a.y, c.y);
      exit;
    end

    else begin               // case 2,3
      StartAndStep (a,b, tc.a, tc.b, x1,d1, c1,f1);
      StartAndStep (a,c, tc.a, tc.c, x2,d2, c2,f2);
      if d1 < d2             // case 2
      then begin
        FillPartFix (a.y, b.y);
        StartAndStep (b,c, tc.b, tc.c, x1,d1, c1,f1);
        FillPartFix (b.y+1, c.y);
      end
      else begin             // case 3
        d3 := d1; d1 := d2; d2 := d3;  // swap d1 with d2
        f3 := f1; f1 := f2; f2 := f3;
        FillPartFix (a.y, b.y);
        StartAndStep (b,c, tc.b, tc.c, x2,d2, c2,f2);
        FillPartFix (b.y+1, c.y);
      end;
    end;
  end;
end;

//*********************************************************
const sBitmapCreation  = 'error at tBitmap32 creation';

initialization
  // create 32 bit work line
  hLine32Bm := TBitmap32.Create;
  if hLine32Bm = nil
//    then Raise EInvalidGraphicsOperation.Create(sBitmapCreation);
  then Halt(255);
  hLine32Bm.SetSize (2048, 11);

//*********************************************************
finalization
  hLine32Bm.Free;

end.

